// MIDI interface class

#include "stdafx.h"
#include "midi.h"
#include "doc.h"
#include "midifish.h"



// shared device lists
static vector<string> gInInterfaces;
static vector<string> gOutInterfaces;



// base class constructor
Midi::Midi(vector<string> &interfaces, Doc *doc)
	: mPort(-1),
	  mChannel(-1),
	  mActive(false),
	  mInterfaces(interfaces),
	  mDoc(doc),
	  mListCtrl(NULL),
	  mEnableCtrl(NULL)
{
}



// base destructor
Midi::~Midi()
{
	ASSERT(!mActive);
}



// give access to UI components
void Midi::init_controls(CComboBox *list_ctrl, CComboBox *channel_ctrl, CButton *enable_ctrl)
{
	mListCtrl = list_ctrl;
	mChannelCtrl = channel_ctrl;
	mEnableCtrl = enable_ctrl;

	mChannelCtrl->ResetContent();

	mChannelCtrl->AddString("*");

	for (int i=1; i<=16; i++) {
		CString s;
		s.Format("%d", i);
		mChannelCtrl->AddString(s);
	}

	mListCtrl->ResetContent();

	mListCtrl->AddString("<none>");

	for (vector<string>::iterator it = mInterfaces.begin(); it != mInterfaces.end(); it++)
		mListCtrl->AddString(it->c_str());

	update_controls();
}



// updates the UI component state
void Midi::update_controls()
{
	mListCtrl->SetCurSel(mPort+1);

	mChannelCtrl->SetCurSel(mChannel+1);

	mEnableCtrl->SetCheck(mActive);
}



// changes the status of a MIDI port
void Midi::set(int port, bool active)
{
	if ((mPort == port) && (mActive == active))
		return;

	if (mActive)
		close();

	mPort = port;

	if (active) {
		if (port >= 0)
			open();

		if (!mActive) {
			CString s;
			s.Format("Unable to open MIDI port\n%s", (mPort >= 0) ? mInterfaces[mPort].c_str() : "<none>");
			AfxMessageBox(s);
		}
	}

	update_controls();

	mDoc->mDirty = true;
}



// reads state from a file
void Midi::load(FILE *file)
{
	int port = fgetl(file);
	int channel = (signed char)fgetc(file);
	int active = fgetc(file);

	if (port >= mInterfaces.size()) {
		port = -1;
		active = false;
	}

	mChannel = channel;

	set(port, active);
}



// saves state to a file
void Midi::save(FILE *file)
{
	fputl(mPort, file);
	fputc(mChannel, file);
	fputc(mActive, file);
}



// input constructor
MidiInput::MidiInput(Doc *doc)
	: Midi(gInInterfaces, doc)
{
	if (mInterfaces.empty()) {
		for (int i=0; i<midiInGetNumDevs(); i++) {
			MIDIINCAPS caps;

			midiInGetDevCaps(i, &caps, sizeof(caps));

			mInterfaces.push_back(caps.szPname);
		}
	}

	mRecentNote = -1;
	mRecentController = -1;
}



// input destructor
MidiInput::~MidiInput()
{
	if (mActive)
		close();
}



// open an input port
void MidiInput::open()
{
	if (mActive)
		close();

	if (mPort < 0)
		return;

	if (midiInOpen(&mHandle, mPort, (unsigned long)callback, (unsigned long)this, CALLBACK_FUNCTION) == MMSYSERR_NOERROR)
		mActive = true;

	midiInStart(mHandle);

	mRecentNote = -1;
	mRecentController = -1;
}



// close an input port
void MidiInput::close()
{
	if (!mActive)
		return;

	mActive = false;

	midiInStop(mHandle);
	midiInClose(mHandle);

	mRecentNote = -1;
	mRecentController = -1;
}



// static member to use as a Windows MIDI input callback
void CALLBACK MidiInput::callback(HMIDIIN handle, UINT msg, DWORD instance, DWORD param1, DWORD param2)
{
	if (msg == MIM_DATA)
		((MidiInput *)instance)->in(param1);
}



// input handler
void MidiInput::in(int data)
{
	int msg = data & 0xF0;
	int channel = data & 0x0F;
	int d1 = (data >> 8) & 0x7F;
	int d2 = (data >> 16) & 0x7F;

	// filter rubbish
	if ((msg == MIDI_PROGRAM_CHANGE) || (msg == MIDI_SYSTEM))
		return;

	if ((mChannel >= 0) && (channel != mChannel))
		return;

	if ((msg == MIDI_NOTE_ON) && (d2 == 0))
		msg = MIDI_NOTE_OFF;

	// store
	switch (msg) {

		case MIDI_NOTE_ON:
			mRecentNote = d1;
			break;

		case MIDI_KEY_PRESSURE:
			mRecentController = CtrlMap::KEY_PRESSURE;
			break;

		case MIDI_CONTROLLER:
			mRecentController = CtrlMap::CONTROLLER + d1;
			break;

		case MIDI_AFTERTOUCH:
			mRecentController = CtrlMap::AFTERTOUCH;
			break;

		case MIDI_PITCH_BEND:
			mRecentController = CtrlMap::PITCH_BEND;
			break;
	}

	// process
	mDoc->process(msg, channel, d1, d2, (this == mDoc->mIn[0]) ? 0 : 1);
}



// output constructor
MidiOutput::MidiOutput(Doc *doc)
	: Midi(gOutInterfaces, doc)
{
	if (mInterfaces.empty()) {
		for (int i=0; i<midiOutGetNumDevs(); i++) {
			MIDIOUTCAPS caps;

			midiOutGetDevCaps(i, &caps, sizeof(caps));

			mInterfaces.push_back(caps.szPname);
		}
	}
}



// output destructor
MidiOutput::~MidiOutput()
{
	if (mActive)
		close();
}



// open an output port
void MidiOutput::open()
{
	if (mActive)
		close();

	if (mPort < 0)
		return;

	if (midiOutOpen(&mHandle, mPort, 0, 0, CALLBACK_NULL) == MMSYSERR_NOERROR)
		mActive = true;
}



// close an output port
void MidiOutput::close()
{
	if (!mActive)
		return;

	mActive = false;

	midiOutClose(mHandle);
}



// ouput MIDI data
void MidiOutput::out(int data)
{
	if (!mActive)
		return;

	if (mChannel >= 0)
		data = (data & ~0xF) | mChannel;

	midiOutShortMsg(mHandle, data);
}



// all notes off
void MidiOutput::panic()
{
	for (int c=0; c<16; c++) {
		out(MIDI_CONTROLLER | c | (123 << 8));		// all notes off
		out(MIDI_CONTROLLER | c | (121 << 8));		// reset all controllers
	}
}
